#include <fcntl.h>
#include <unistd.h>
#include <getopt.h>
#include <errno.h>
#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <inttypes.h>
#include <dirent.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/Xproto.h>
#include <X11/keysym.h>
#include <X11/Xmu/WinUtil.h>

#include "scanner.h"
#include "data.h"
#include "panic.h"
#include "xatoms.h"
#include "mod.h"
#include "print_event.h"

#define NUM(x) (sizeof(x)/sizeof(x[0]))

static const char *files[] = {
        ".emap.d",
        "etc/emap.d"
};


static int grabkey_errors;
static bool foreground = false;
static char *logfile = NULL;


void ParseTrace(FILE *TraceFILE, char *zTracePrompt);
static int parse_file(char *file, rb_t *rb);
static int map_compare(const void *a, const void *b);
static void run_emap(int argc, char **argv, char *, struct map *maps, int nmaps);
static void do_kdump(int kmode);
static int error_handler(Display *, XErrorEvent *);
static void setenvf(char *name, char *value, ...);

int
main(int argc, char *argv[])
{
        bool only_listed = false;
        char *display = NULL;
        int errors = 0;
        rb_t rb = RB_BLANK;
        int ch;
        int kmode = 0;
        while ((ch = getopt(argc, argv, "fvkKod:l:")) != -1) switch(ch) {
        case 'd':
                display = optarg;
                break;
        case 'o':
                only_listed = true;
                break;
        case 'f':
                foreground = true;
                break;
        case 'v':
                verbose_mode = true;
                foreground = true;
                break;
        case 'l':
                logfile = optarg;
                break;
                
        case 'k':
        case 'K':
                kmode = ch;
                break;
        default:
                panicf("unrecognized option '%c'", ch);
                break;
        }
        if (kmode) {
                do_kdump(kmode);
                return 0;
        }

        if (verbose_mode) 
                ParseTrace(stderr, "> ");
        for(;optind < argc; optind++)
                errors += parse_file(argv[optind], &rb);
        char *home = getenv("HOME");
        if(!only_listed && home) {
                int i;
                size_t sz_h = strlen(home);
                for(i = 0; i < NUM(files); i++) {
                        DIR *dir;
                        struct dirent *dp;
                        size_t sz_f = strlen(files[i]);
                        char name[sz_h + sz_f + 2];
                        strcpy(name, home); 
                        strcat(name, "/");
                        strcat(name, files[i]);
                        dir = opendir(name);
                        print_verbose("%s: %s", name, dir ? "reading" : "could not open as directory");
                        if(!dir) 
                                continue;
                        while ((dp = readdir(dir))) {
                                size_t len = strlen(dp->d_name);
                                if (len > 5 && !memcmp(dp->d_name + len - 5, ".emap", 5)) {
                                        char nname[sz_h + sz_f + len + 3];
                                        strcpy(nname, name);
                                        strcat(nname, "/");
                                        strcat(nname, dp->d_name);
                                        errors += parse_file(nname, &rb);
                                }
                        }
                }

        }
        if(errors > 0) {
                if (errors == 1)
                        panic("There was an error.");
                else
                        panicf("There were %i errors.", errors);
        }
        assert(RB_LEN(&rb) % sizeof(struct map) == 0);

        run_emap(argc, argv, display, RB_PTR(&rb), RB_LEN(&rb)/sizeof(struct map));

        return 0;
}

static void
setenvf(char *name, char *value, ...)
{
        char *v;
        va_list ap;
        va_start(ap, value);
        vasprintf( &v, value, ap);
        assert(v);
        va_end(ap);
        setenv(name, v, 1);
        free(v);
}

static void
fork_away(void)
{
        pid_t pid;
        int fd;
        int logfd = -1;
        if(logfile) {
                logfd = open(logfile, O_APPEND | O_WRONLY | O_CREAT);
                if(logfd == -1) {
                        perror(logfile);
                        fputs("Could not open log file.\n", stderr);
                        exit(1);
                }
        }
        fflush(stdout);
        fflush(stderr);
        if ((pid = fork()) < 0) {
                perror("fork");
                fputs("Running in foreground...\n", stderr);
                return;
        }
        if (pid != 0) exit(0);
        setsid();
        close(0);
        close(1);
        close(2);
        fd = open("/dev/null", O_RDWR);
        fd = logfd == -1 ? fd : logfd;
        dup(fd);
        dup(fd);
}


static int 
map_compare(const void *aa,const void *bb)
{
        int ret;
        const struct map *a = aa;
        const struct map *b = bb;
        //ret = strcmp(a->leader.name, b->leader.name);
        ret = a->leader.keysym - b->leader.keysym;
        if(ret)
                return ret; 
        if(a->leader.state == AnyModifier || b->leader.state == AnyModifier) 
                return strcmp(a->string, b->string);
        ret = a->leader.state - b->leader.state;
        if(ret)
                return ret; 
        return strcmp(a->string, b->string);
}

/*
 * returns -1 if a is submap of b
 * 0 if a == b
 * 1 if a and b are incomparable
 */
static int 
map_submap(struct key leader, char *string, const struct map *b)
{
        int la, lb, ret;
        ret = leader.keysym - b->leader.keysym || (leader.state - b->leader.state && b->leader.state != AnyModifier)  ;
        if(ret)
                return 1; 
        la = strlen(string);
        lb = strlen(b->string);
        if (la > lb)
                return 1;
        if(!memcmp(string, b->string, la)) {
                if(la == lb)
                        return 0;
                else 
                        return -1;
        }
        return 1;
        
}        

static int 
parse_file(char *file, rb_t *rb) 
{
        struct parse_data pd;
        int fd;
        print_verbose("Parsing file '%s'...", file);
        memset(&pd, 0, sizeof(pd));
        if(!setjmp(pd.error_jmp)) {
                pd.filename = file;
                fd = open(file, O_RDONLY);
                if(fd == -1) {
                        perror(file);
                        return 1;
                }
                pd.rb = *rb;
                scan_emap(fd, &pd);
                *rb = pd.rb;
                print_verbose("%s parsed. errors: %i", file, pd.errorcount);
                return pd.errorcount;
        } else 
                return pd.errorcount;
}

static void
grab_hotkeys(Display *dpy, struct map *maps, int nmaps)
{
        int i,j;
        struct key last = {NULL, 0, 0};
        assert (nmaps >= 0);


        XSetErrorHandler(error_handler);
        XSync(dpy, False);

        grabkey_errors = 0;

        for(i = 0; i < nmaps; i++) {
                if (!memcmp(&maps[i].leader,&last,sizeof(last)))
                        continue;
                last = maps[i].leader;


                for (j = 0; j < ScreenCount(dpy); j++) {
                        XGrabKey(dpy, XKeysymToKeycode(dpy, maps[i].leader.keysym), maps[i].leader.state, RootWindow(dpy, j), False, GrabModeAsync, GrabModeSync);
                }
        }
        XSync(dpy, False);
        XSetErrorHandler(NULL);
        if(grabkey_errors) {
                fprintf(stderr, "There were errors grabbing some of the keys requested, perhaps another app is using them.\n");
        }



}

#define MAX_BUF  256

static void
xkey_to_key(XKeyEvent *xkey, struct key *key)
{
        key->keysym = XLookupKeysym(xkey, 0);
        key->state = xkey->state & ~ignored_mask;
}

static void
do_system(char *what, struct map *map, XKeyEvent *xkey)
{
        char *disp = malloc(strlen(DisplayString(xkey->display)) + 20);
        char *c;
        int dummy, i;
        int screen = 0;
        XClassHint classHint;
        Window window, app_window;
        XGetInputFocus(xkey->display, &window, &dummy);
        if(window == PointerRoot || window == None) window = xkey->subwindow;
        app_window = XmuClientWindow(xkey->display, window);
        for(i = 0; i < ScreenCount(xkey->display); i++) 
                if(RootWindow(xkey->display, i) == xkey->root) 
                        screen = i;
        strcpy(disp, DisplayString(xkey->display));
        if ((c = strrchr (disp, ':'))) {
                char *dot;
                dot = strrchr(c, '.');
                if (!dot)
                        dot = c + strlen(c);
                sprintf(dot, ".%i", screen);
        }



        setenv("DISPLAY", disp, 1);
        setenvf("E_ROOT", "%u", xkey->root);
        setenvf("E_MOUSE_WINDOW", "%u", xkey->subwindow);
        setenvf("E_WINDOW", "%u", window);
        setenvf("E_SCREEN", "%u", screen);
        setenvf("E_APP_WINDOW", "%u", app_window);
        setenvf("E_KEY", "%s", map->leader.name);
        setenvf("E_STRING", "%s", map->string);
        setenvf("E_MOUSE_COORD", "%u,%u", xkey->x_root, xkey->y_root);
        if(XGetClassHint(xkey->display, app_window, &classHint)) {
                setenvf("E_APP_CLASS", "%s/%s", classHint.res_name, classHint.res_class);
                XFree(classHint.res_name);
                XFree(classHint.res_class);
        } else
                unsetenv("E_APP_CLASS");

        //TODO E_APP_NAME, E_APP_COMMAND, E_APP_PID, E_APP_CLIENT_MACHINE
        system(what);
}

static void
execute_action(struct map *map, XKeyEvent *xkey)
{
        print_verbose("Executing action: ");
        if (verbose_mode)
                dump_maps(stderr, map, 1);
        switch(map->atype) {
        case ACTION_EXEC:
                do_system(map->adata, map, xkey);
                break;
        case ACTION_INJECT:
                //TODO
                break;
        case ACTION_EXIT:
                exit(0);
        case ACTION_DIE:
                exit(1);
        }
}

static int
error_handler(Display *dpy, XErrorEvent *ev)
{
        if(ev->request_code == X_GrabKey) {
                grabkey_errors++;
        } 
        return 0;
}

static void
run_emap(int argc, char **argv, char *display, struct map *maps, int nmaps)
{
        int i;
        Display *dpy;
        XEvent xev;
        char buf[MAX_BUF];
        int ibuf = 0;
        struct key leader;
        memset(&leader, 0, sizeof leader);


        for(i = 0; i < nmaps; i++) 
                maps[i].leader.state &= ~ignored_mask;

        qsort(maps, nmaps, sizeof(struct map), map_compare);
        if(verbose_mode)
                dump_maps(stderr, maps, nmaps);

        dpy = XOpenDisplay(display);
        if(!dpy) 
                panicf("Could not open display: %s", XDisplayName(display));
        intern_xatoms(dpy);


        grab_hotkeys(dpy, maps, nmaps);

        if(!foreground)
                fork_away();


        for (;;) {
                int status;
                XNextEvent(dpy, &xev);
                if(verbose_mode)
                        print_event( xev);
                switch(xev.type) {
                case KeyPress:
                        if(!leader.keysym) {
                                XGrabKeyboard(dpy, DefaultRootWindow(dpy), False, GrabModeAsync, GrabModeSync, xev.xkey.time);
                                print_verbose("Grabbing keyboard\n");
                                //XAllowEvents(dpy, SyncKeyboard, CurrentTime);
                                xkey_to_key(&xev.xkey, &leader);
                                ibuf = 0;
                                buf[0] = '\0';
                        } else {
                                int sz;
                                sz = XLookupString(&xev.xkey, buf + ibuf, MAX_BUF - ibuf - 1, NULL, NULL);
                                ibuf += sz;
                                buf[ibuf] = '\0';
                        }
                        status = 0;
                        for(i = 0; i < nmaps; i++) {
                                int st = map_submap(leader, buf, &maps[i]);
                                if(!st) {
                                        execute_action(&maps[i], &xev.xkey);
                                        XAllowEvents(dpy, AsyncKeyboard, CurrentTime); 
                                        XUngrabKeyboard(dpy, xev.xkey.time);
                                        memset(&leader, 0, sizeof leader);
                                        break;
                                }
                                if(st < 0) status = 1;
                        }
                        if(i == nmaps) {
                                if(status) { //still possibility of matching something
                                        XAllowEvents(dpy, SyncKeyboard, CurrentTime); 
                                } else {
                                        XAllowEvents(dpy, ReplayKeyboard, CurrentTime); 
                                        XUngrabKeyboard(dpy, xev.xkey.time);
                                        memset(&leader, 0, sizeof leader);
                                }
                        }




                        if (XK_q ==  XLookupKeysym(&xev.xkey, 0)) {

                                XUngrabKeyboard(dpy, xev.xkey.time);
                                buf[ibuf] = '\0';
                                printf("%s\n", buf);
                                memset(&leader, 0, sizeof leader);
                                break;
                        }

                case KeyRelease:
                        XAllowEvents(dpy, SyncKeyboard, CurrentTime);
                        break;
                case MappingNotify:
                        print_verbose("MappingNotify: regrabing hotkeys...\n");
                        for(i = 0; i < ScreenCount(dpy); i++) {
                                XUngrabKey(dpy, AnyKey, AnyModifier, RootWindow(dpy, i));
                        }
                        XRefreshKeyboardMapping(&xev.xmapping);
                        grab_hotkeys(dpy, maps, nmaps);
                        break;

                default:
                        printf("unknown X event.\n");
                        print_event( xev);

                }

        }



}

static void
do_kdump(int kmode)
{
        XEvent xev;
        struct key key;
        Display *dpy = XOpenDisplay(NULL);
        if(!dpy) 
                panic("Could not open display");
        printf("Showing keyevents. press 'q' to quit...\n");
        XGrabKeyboard(dpy, DefaultRootWindow(dpy), False, GrabModeAsync, GrabModeAsync, CurrentTime);
        for (;;) {
                XNextEvent(dpy, &xev);
                switch(xev.type) {
                case KeyPress:
                        if(verbose_mode)
                                print_event( xev);
                        key.keysym = XLookupKeysym(&xev.xkey, 0);
                        key.state = xev.xkey.state;
                        printf("<");
                        print_key(stdout, &key);
                        printf(">\n");
                        if (XK_q == key.keysym) 
                                return;

                case KeyRelease:
                        break;
                case MappingNotify:
                        XRefreshKeyboardMapping(&xev.xmapping);
                        break;
                default:
                        printf("unknown X event.\n");
                        print_event( xev);

                }

        }


}


